// Kevin Beason
// http://www.cs.fsu.edu/~beason/
// 4/17/02
// 
// This program implements a simple Photon Mapping algorithm
// using Open Inventor. It requires a decimated brain MRI data set
// if you use the -b option (b for brain).
// 
// It writes out a scene file every so often. You can also save it at any
// time by pressing the S key. Also, you can press B to turn the faceset
// on/off, and P to turn the photon map on/off.
// 
// Some code is from:
// 
// Josh Grant (http://www.cs.fsu.edu/~grant/)
// Dr. David C. Banks (http://www.cs.fsu.edu/~banks/)

#include <math.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <string.h>
#include <stdio.h>

#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/SoOutput.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoCone.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoMaterialBinding.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/sensors/SoTimerSensor.h>
#include <Inventor/nodes/SoComplexity.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoDB.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/SbLinear.h>
#include <Inventor/nodes/SoShapeHints.h>
#include <Inventor/engines/SoCalculator.h>
#include <Inventor/nodes/SoVertexProperty.h>
#include <Inventor/nodes/SoIndexedFaceSet.h>
#include <Inventor/fields/SoSubField.h>
#include <Inventor/SbLinear.h>
#include <Inventor/nodes/SoIndexedLineSet.h>
#include <Inventor/events/SoKeyboardEvent.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/nodes/SoPointSet.h>
#include <Inventor/nodes/SoSwitch.h>

#include "MarchingCubes.h"
#include "intersect.h"

int numVertices = 0;
int verticesPerPass = 1;
int raysPerVertex = 60;
double myGamma = 0.15;
SoCamera *Camera;
SoIndexedFaceSet *faceSet; // (SoMFInt32 coordIndex)
SoVertexProperty *vprop;   // (SoMFVec3f vertex, SoMFVec3f normal)
SoMaterial *faceMat;
SoSeparator *root;
SoSwitch *brainSwitch;
SoSwitch *photSwitch;

const int valSize = 1;
const int maxDim = 256;
const int backDim[2] = {1, 1};
int useBrain = 0;
int useWindow = 1;
int dataHeaderSize = 0;
char dataFileName[] = "brain2";
int dims[3] = { 28, 28, 28 };
float data[28*28*28];
SoMFVec3f intens; // intensities 0 = dark 10000 = sun
SoMFVec3f intens2; // temporary intensity storage
const float sun_intens = 10000.0;

typedef struct rayStruct {
  SbVec3f orig;
  SbVec3f dir;
} rayType;

SoMFVec3f *verts;
SoMFVec3f *norms;
SoMFInt32 *indxs;
int firstLightVert;
int lastLightVert;

unsigned int
readVal(unsigned char *ptr, int valSize, int bigEndian=0){
  int i;
  unsigned int c = 0;

  if (bigEndian){
    for (i=valSize-1; i>=0; i--){
      c = (c << 8) | ptr[i];
    }
  }
  else {
    for (i=0; i<valSize; i++){
      c = (c << 8) | ptr[i];
    }
  }

  // tweak for stanford bunny
  if (c == 0xf830)
    c = 0;
  
  return c;
}

void
readScalarField(){

  // open file in binary
 
  FILE *dataFile = fopen(dataFileName, "rb");
  if (!dataFile){
    cerr << "Could not open file: " << dataFileName << endl;
    exit(1);
  }

  // skip header
  
  int c;
  unsigned char buffer[maxDim*maxDim*valSize];
  int maxc=0, minc=9999999;
  
  // iterate through each grid point
  for (int k = 0; k < dims[2]; k++) {

    // read in data
    cerr << "\rReading in slice " << k+1 << "/" << dims[2];
    fread(buffer, valSize, dims[0]*dims[1], dataFile);
    
    for (int j = 0; j < dims[1]; j++) {
      for (int i = 0; i < dims[0]; i++) {

	// compute data
	c = readVal(&buffer[(j*dims[0]+i)*valSize], valSize, 0);
	if (c<minc)
	  minc = c;
	else if (c>maxc)
	  maxc = c;
	
	data[k*dims[1]*dims[0] + j*dims[0]+i] = (float)(123-c);

      }
    }
  }
  cout << endl;
  cout << "Max: " << maxc << "  "
       << "Min: " << minc << endl;
  fclose(dataFile);
}

const SbVec3f min = SbVec3f(-1.0, -1.0, -1.0);
const SbVec3f maxDiff = SbVec3f( 2.0, 2.0, 2.0);
const SbVec3f max = min + maxDiff;

// Non-linear scale
const float gamma_val = 1000.0;
inline float nlscale(const float &x){
  return(x/gamma_val/(1.0+x/gamma_val));
}

// Non-linear scale
const inline SbColor nlscale(const SbVec3f x){
  return(SbColor(x[0]/gamma_val/(1.0+x[0]/gamma_val),
		 x[1]/gamma_val/(1.0+x[1]/gamma_val),
		 x[2]/gamma_val/(1.0+x[2]/gamma_val) ));
}

class LineSet {
private:
  SoCoordinate3 *coords;
  SoIndexedLineSet *index;
  SoSeparator *sep;
  SoMaterial *mat;
  SoMaterialBinding *bind;
  int numCoords, numIndexes;
  
public:
  LineSet(){
    coords = new SoCoordinate3();
    index = new SoIndexedLineSet();
    bind = new SoMaterialBinding();
    bind->value = SoMaterialBinding::PER_VERTEX_INDEXED;
    mat = new SoMaterial();
    sep = new SoSeparator();
    mat->diffuseColor.set1Value(0, SbColor(1.0, 0.0, 0.0)); // red
    mat->diffuseColor.set1Value(1, SbColor(1.0, 1.0, 1.0)); // white
    coords->point.set1Value(0, SbVec3f(0.0, 0.0, 0.0));
    coords->point.set1Value(1, SbVec3f(0.0, 0.0, 0.0));
    index->coordIndex.set1Value(0, 0);
    index->coordIndex.set1Value(1, 1);
    index->coordIndex.set1Value(2, -1);
    numCoords = 2;
    numIndexes = 3;
    sep->addChild(bind);
    sep->addChild(mat);
    sep->addChild(coords);
    sep->addChild(index);
  }

  SoSeparator *getSep(){
    return(sep);
  }
      
  void draw(const SbVec3f &a, const SbVec3f &b){
    // set coords
    coords->point.set1Value(numCoords++, a);
    coords->point.set1Value(numCoords++, b);

    // set index
    index->coordIndex.set1Value(numIndexes++, numCoords-2);
    index->coordIndex.set1Value(numIndexes++, numCoords-1);
    index->coordIndex.set1Value(numIndexes++, -1);

    // set color
    mat->diffuseColor.set1Value(numCoords-2, SbColor(0.0, 0.0, 1.0));
    mat->diffuseColor.set1Value(numCoords-1, SbColor(1.0, 1.0, 0.0));
  }

  void draw(const SbVec3f &a, const SbVec3f &b, const SbColor &c){
    // set coords
    coords->point.set1Value(numCoords++, a);
    coords->point.set1Value(numCoords++, b);

    // set index
    index->coordIndex.set1Value(numIndexes++, numCoords-2);
    index->coordIndex.set1Value(numIndexes++, numCoords-1);
    index->coordIndex.set1Value(numIndexes++, -1);

    // set color
    mat->diffuseColor.set1Value(numCoords-2, SbColor(0.0, 1.0, 1.0));
    mat->diffuseColor.set1Value(numCoords-1, c);
  }

  void draw1(const SbVec3f &a, const SbVec3f &b, const SbColor &c){

    // set coords
    coords->point.set1Value(0, a);
    coords->point.set1Value(1, b);

    // set color
    mat->diffuseColor.set1Value(1, c);
  }

};

class PointSet {
private:
  SoSeparator *sep;
  SoCoordinate3 *coord;
  SoMaterial *mat;
  SoPointSet *ps;
  int numpts;
  
public:
  PointSet(){
    sep = new SoSeparator();
    coord = new SoCoordinate3();
    mat = new SoMaterial();
    mat->ambientColor.setValue(0.0, 0.0, 0.0);
    mat->diffuseColor.setValue(0.0, 0.0, 0.0);
    mat->specularColor.setValue(0.0, 0.0, 0.0);
    mat->emissiveColor.setValue(0.0, 0.0, 0.0);
    mat->shininess.setValue(0.0);
    mat->transparency.setValue(0.0);
    ps = new SoPointSet();
    ps->vertexProperty = coord;
    ps->startIndex = 0;
    ps->numPoints = 0;
    sep->addChild(mat);
    sep->addChild(coord);
    sep->addChild(ps);
    numpts = 0;
  }

  SoSeparator *getSep(){
    return(sep);
  }

  int getNum(){
    return(numpts);
  }

  void add(SbVec3f p, SbColor c = SbColor(1.0, 1.0, 1.0)){
    coord->point.set1Value(numpts, p);
    mat->diffuseColor.set1Value(numpts, c);
    ps->numPoints = numpts+1; 
    numpts++;
  }
  
};

LineSet *lineSet;
PointSet *pointSet;

void doLighting(){

  // Loop through verts
  int vert;

  for (vert = 0; vert<=vprop->vertex.getNum(); vert++){
    
  }
    
  
}

void writeScene(){
  static int sceneNum = 0;
  char filename[50];
  SoOutput *output;
  SoWriteAction *wa;

  // write out scene
  lineSet->draw1(vprop->vertex[0],
		 vprop->vertex[0],
		 SbColor(0.0, 0.0, 0.0));

  output = new SoOutput();
  sprintf(filename, "scenes/scene%d.iv", sceneNum++);
  if(output->openFile(filename)){
    wa = new SoWriteAction(output);
    wa->apply(root);
    output->closeFile();
  }
  else {
    fprintf(stderr, "unknown file i/o error :(\n");
  }
}


//======================================================================
// This key press code is from Josh Grant (http://www.cs.fsu.edu/~grant)
//======================================================================
void keyboardEvents (void *data, SoEventCallback *eventCB)
{
  const SoEvent *ev = eventCB->getEvent();

  // Turn off brain with B key!
  if (SO_KEY_PRESS_EVENT(ev, B))
    brainSwitch->whichChild.setValue(1-brainSwitch->whichChild.getValue());

  // Turn off photons with P key!
  else if (SO_KEY_PRESS_EVENT(ev, P))
    photSwitch->whichChild.setValue(1-photSwitch->whichChild.getValue());

  // Calculate Vertex lighting with the L key!
  else if (SO_KEY_PRESS_EVENT(ev, L))
    doLighting();

  // Calculate Vertex lighting with the L key!
  else if (SO_KEY_PRESS_EVENT(ev, S)){
    cerr << "Writing scene.\n";
    writeScene();
  }
  
}


SoSeparator *
makeScene (){

  SoSeparator *scene = new SoSeparator();


  // This keypress code is from Josh Grant
  SoEventCallback *cb = new SoEventCallback();
  cb->addEventCallback(SoKeyboardEvent::getClassTypeId(),
                       keyboardEvents, NULL);
  scene->addChild(cb);

  
  // Cube
  SoSeparator *cubeSep = new SoSeparator();
  SoMaterial *cubeMat = new SoMaterial();
  cubeMat->diffuseColor.setValue(1.0, 0.3, 1.0);
  cubeSep->addChild(cubeMat);
  
  /*     ____
       /|1  /|5    z
      / |  / |     |
     /__|0=min4    |__y
    2|  / |6=max  /
     | /  | /    x
    3|/___|7

  */
    
  static float vertices[8][3] =
  {  { min[0], min[1], min[2]},   // 0
     { min[0], min[1], max[2]},   // 1
     { max[0], min[1], max[2]},   // 2
     { max[0], min[1], min[2]},   // 3
     { min[0], max[1], min[2]},   // 4
     { min[0], max[1], max[2]},   // 5
     { max[0], max[1], max[2]},   // 6
     { max[0], max[1], min[2]}    // 7
  };

  static int32_t indices[40] =
  {
    0, 1, 2, 3,  -1,
    2, 6, 7, 3,  -1,
    4, 7, 6, 5,  -1,
    0, 4, 5, 1,  -1,
    1, 5, 6, 2,  -1,
    0, 3, 7, 4,  -1
  };

  // coordinates
  SoCoordinate3 *cubeCoords = new SoCoordinate3();
  cubeCoords->point.setValues(0,8, vertices);
  cubeSep->addChild(cubeCoords);

  // line set
  SoIndexedLineSet *cubeLineSet = new SoIndexedLineSet();
  cubeLineSet->coordIndex.setValues(0, 40, indices);
  cubeSep->addChild( cubeLineSet );

  // lineset class
  lineSet = new LineSet();
  scene->addChild(lineSet->getSep());

  // pointset class and switch
  pointSet = new PointSet();
  photSwitch = new SoSwitch();
  photSwitch->addChild(pointSet->getSep());
  photSwitch->addChild(new SoSeparator());
  photSwitch->whichChild.setValue(0);
  scene->addChild(photSwitch);

  // add to scene
  scene->addChild(cubeSep);

////////////////////////////////////////////////////////////////////////
// Begin Josh Grant's code
// http://www.cs.fsu.edu/~grant/
////////////////////////////////////////////////////////////////////////

  // create MarchingCubes engine and connect fields
  MarchingCubes *mcubes = new MarchingCubes();
  mcubes->ref();

  // set the data
  readScalarField();
  mcubes->data.setValue(dims, data, false);

  // set isoValue, in this case this means find the surface of the sphere
  // where the radius is 1.0
  mcubes->isoValue = 70.0;

  // MarchingCubes gives the triangles with the vertices ordered clockwise
  SoShapeHints *hints = new SoShapeHints();
  hints->vertexOrdering.setValue(SoShapeHints::CLOCKWISE);
  hints->shapeType.setValue(SoShapeHints::SOLID);
  scene->addChild(hints);

  // Create a VertexProperty because they are more efficient the
  // SoCoordinate3 nodes.
  vprop = new SoVertexProperty();
  vprop->ref();
  vprop->vertex.connectFrom(&mcubes->points);
  vprop->normal.connectFrom(&mcubes->normals);

  // connect the indexes from MarchingCubes
  SoSeparator *faceSep = new SoSeparator();
  faceMat = new SoMaterial();
  SoMaterialBinding *faceMatBind = new SoMaterialBinding();
  faceMatBind->value = SoMaterialBinding::PER_VERTEX_INDEXED;
  faceMat->ambientColor.setValue(0.0, 0.0, 0.0);
  faceMat->diffuseColor.setValue(0.5, 0.5, 0.5);
  faceMat->specularColor.setValue(0.0, 0.0, 0.0);
  faceMat->emissiveColor.setValue(0.0, 0.0, 0.0);
  faceMat->shininess.setValue(0.0);
  faceMat->transparency.setValue(0.0);
  faceSet = new SoIndexedFaceSet();
  faceSet->vertexProperty = vprop;
  faceSet->coordIndex.connectFrom(&mcubes->indexes);
  faceSep->addChild(faceMatBind);
  faceSep->addChild(faceMat);
  faceSep->addChild(faceSet);

  // add switch
  brainSwitch = new SoSwitch();
  brainSwitch->addChild(faceSep);
  brainSwitch->addChild(new SoSeparator());
  brainSwitch->whichChild.setValue(1);
  scene->addChild(brainSwitch);

////////////////////////////////////////////////////////////////////////  
// End Josh Grant's code
////////////////////////////////////////////////////////////////////////

  // disconnect so i can relax
  vprop->vertex.disconnect();
  vprop->normal.disconnect();
  faceSet->coordIndex.disconnect();

  // set my global vars
  verts = &(vprop->vertex);
  norms = &(vprop->normal);
  indxs = &(faceSet->coordIndex);

  // kill the brain!
  if (! useBrain){
    vprop->vertex.setNum(0);
    vprop->normal.setNum(0);
    faceSet->coordIndex.setNum(0);

    // instead, put a triangle in
    vprop->vertex.set1Value(0, SbVec3f(min[0]+maxDiff[0]*2.0/3.0,
				       min[1]+maxDiff[1]/3.0,
				       min[2]+maxDiff[2]/3.0));
    vprop->vertex.set1Value(1, SbVec3f(min[0]+maxDiff[0]*1.0/3.0,
				       min[1]+maxDiff[1]/3.0,
				       min[2]+maxDiff[2]/3.0));
    vprop->vertex.set1Value(2, SbVec3f(min[0]+maxDiff[0]*1.0/3.0,
				       min[1]+maxDiff[1]*2.0/3.0,
				       min[2]+maxDiff[2]*2.0/3.0));
    vprop->normal.set1Value(0, SbVec3f(0.0, -1.0/sqrt(2.0), 1.0/sqrt(2.0)));
    vprop->normal.set1Value(1, SbVec3f(0.0, -1.0/sqrt(2.0), 1.0/sqrt(2.0)));
    vprop->normal.set1Value(2, SbVec3f(0.0, -1.0/sqrt(2.0), 1.0/sqrt(2.0)));
    faceSet->coordIndex.set1Value(0, 0);
    faceSet->coordIndex.set1Value(1, 1);
    faceSet->coordIndex.set1Value(2, 2);
    faceSet->coordIndex.set1Value(3, -1);

  }


  // add back face polygons
  int i, j;
  int vi, vi_orig, ci;

  // create coordinates and normals
  vi = vprop->vertex.getNum();
  vi_orig = vi;
  for (j=0; j<=backDim[1]; j++)
    for (i=0; i<=backDim[0]; i++){
      vprop->vertex.set1Value(vi, SbVec3f(min[0]+maxDiff[0]*i/
					  (float)backDim[0],
					  max[1],
					  min[2]+maxDiff[2]*j/
					  (float)backDim[1]));
      vprop->normal.set1Value(vi, SbVec3f(0.0, -1.0, 0.0));
      vi++;
    }

  // create faces
  ci = faceSet->coordIndex.getNum();
  vi = vi_orig;
  for (j=0; j<backDim[1]; j++){
    for (i=0; i<backDim[0]; i++){
      faceSet->coordIndex.set1Value(ci+0, vi);
      faceSet->coordIndex.set1Value(ci+1, vi+backDim[0]+1);
      faceSet->coordIndex.set1Value(ci+2, vi+1);
      faceSet->coordIndex.set1Value(ci+3, -1);
      faceSet->coordIndex.set1Value(ci+4, vi+1);
      faceSet->coordIndex.set1Value(ci+5, vi+backDim[0]+1);
      faceSet->coordIndex.set1Value(ci+6, vi+backDim[0]+2);
      faceSet->coordIndex.set1Value(ci+7, -1);
      ci+=8;
      vi++;
    }
    vi++;
  }

  // add bottom face polygons

  // create coordinates and normals
  vi = vprop->vertex.getNum();
  vi_orig = vi;
  for (j=0; j<=backDim[1]; j++)
    for (i=0; i<=backDim[0]; i++){
      vprop->vertex.set1Value(vi, SbVec3f(min[0]+maxDiff[0]*i/
					  (float)backDim[0],
					  min[1]+maxDiff[1]*j/
					  (float)backDim[1],
					  min[2]));
      vprop->normal.set1Value(vi, SbVec3f(0.0, 0.0, 1.0));
      vi++;
    }

  // create faces
  ci = faceSet->coordIndex.getNum();
  vi = vi_orig;
  for (j=0; j<backDim[1]; j++){
    for (i=0; i<backDim[0]; i++){
      faceSet->coordIndex.set1Value(ci+0, vi);
      faceSet->coordIndex.set1Value(ci+1, vi+backDim[0]+1);
      faceSet->coordIndex.set1Value(ci+2, vi+1);
      faceSet->coordIndex.set1Value(ci+3, -1);
      faceSet->coordIndex.set1Value(ci+4, vi+1);
      faceSet->coordIndex.set1Value(ci+5, vi+backDim[0]+1);
      faceSet->coordIndex.set1Value(ci+6, vi+backDim[0]+2);
      faceSet->coordIndex.set1Value(ci+7, -1);
      ci+=8;
      vi++;
    }
    vi++;
  }

  // create light
  // set up center, center offset, normal
  const float sqd = 0.5;
  SbVec3f c = SbVec3f(0.0, -2.0, 2.0);
  SbVec3f coff = SbVec3f(-sqd*sqrt(2.0), -sqd, -sqd);
  //  SbVec3f n = -coff.cross(SbVec3f(coff[0], coff[1], -coff[2]));
  SbVec3f n = -c;
  n.normalize();
			 
  // create coordinates
  vi = vprop->vertex.getNum();
  ci = faceSet->coordIndex.getNum();
  firstLightVert = vi_orig = vi;
  lastLightVert = firstLightVert+3;
  
  vprop->vertex.set1Value(vi++, c+coff);
  vprop->vertex.set1Value(vi++, SbVec3f(c[0]+coff[0],
					c[1]-coff[1],
					c[2]-coff[2]));
  vprop->vertex.set1Value(vi++, c-coff);
  vprop->vertex.set1Value(vi++, SbVec3f(c[0]-coff[0],
					c[1]+coff[1],
					c[2]+coff[2]));
  
  // set normals
  vi = vi_orig;
  vprop->normal.set1Value(vi++, n);
  vprop->normal.set1Value(vi++, n);
  vprop->normal.set1Value(vi++, n);
  vprop->normal.set1Value(vi++, n);

  // create faces
  faceSet->coordIndex.set1Value(ci++, vi_orig+0);
  faceSet->coordIndex.set1Value(ci++, vi_orig+3);
  faceSet->coordIndex.set1Value(ci++, vi_orig+1);
  faceSet->coordIndex.set1Value(ci++, -1);
  faceSet->coordIndex.set1Value(ci++, vi_orig+1);
  faceSet->coordIndex.set1Value(ci++, vi_orig+3);
  faceSet->coordIndex.set1Value(ci++, vi_orig+2);
  faceSet->coordIndex.set1Value(ci++, -1);
  

  // set intensities
  for (i = 0; i < vprop->vertex.getNum(); i++)
    intens.set1Value(i, SbVec3f(0.0, 0.0, 0.0));
  // turn light on
  for (i = vi_orig; i < vi_orig+4; i++)
    intens.set1Value(i, SbVec3f(sun_intens/1.0,
				sun_intens/1.0,
				sun_intens/1.0));
  // set intensities2
  for (i = 0; i < vprop->vertex.getNum(); i++)
    intens2.set1Value(i, intens[i]);

  // set colors
  for (i = 0; i < vprop->vertex.getNum(); i++)
    faceMat->diffuseColor.set1Value(i, SbVec3f(nlscale(intens[i][0]),
					       nlscale(intens[i][1]),
					       nlscale(intens[i][2]) ));

  // set number of vertices
  numVertices = vprop->vertex.getNum();

  return scene;
}

void debug_showray(int currVertex, int hit_idx, SbVec3f dpn){
	  
  faceMat->diffuseColor.set1Value(faceSet->coordIndex[hit_idx],
				  SbColor(0.5, 0.0, 0.0));
  faceMat->diffuseColor.set1Value(faceSet->coordIndex[hit_idx+1],
				  SbColor(0.0, 0.5, 0.0));
  faceMat->diffuseColor.set1Value(faceSet->coordIndex[hit_idx+2],
				  SbColor(0.0, 0.0, 0.5));
  faceMat->diffuseColor.set1Value(currVertex,
				  faceMat->diffuseColor[currVertex] +
				  SbColor(0.5, 0.5, 0.5));

  SoMaterial *myMat = new SoMaterial();
  myMat->diffuseColor.setValue(SbColor(1.0, 1.0, 1.0));
  SoTransform *myXform = new SoTransform();
  SoTransform *myXform2 = new SoTransform();
  myXform->translation = vprop->vertex[currVertex];
  myXform2->translation=0.1*dpn;
  SoSphere *myS = new SoSphere();
  SoSphere *myS2 = new SoSphere();
  myS->radius.setValue(0.01);
  myS2->radius.setValue(0.01);
  root->addChild(myMat);
  root->addChild(myXform);
  root->addChild(myS);
  root->addChild(myXform2);
  root->addChild(myS2);

}







void fireray(rayType &ray, int visRay = 0){

  int vertex = 0;
  int i;
  double orig[3], dir[3], vert0[3], vert1[3], vert2[3];
  double t, u, v, last_t, hit_t;
  int idx, hit_idx, hit_vtx, hit_vtx0, hit_vtx1, hit_vtx2, numIndexes;
  SbRotation sbr;
  SbVec3f dp, dpn;
  double phi, theta;

  numIndexes = indxs->getNum();
  last_t = HUGE;
  hit_idx = -1;
      
  // loop through indexes (triangles) to find closest triangle
  for (idx=0; idx<numIndexes; idx+=4){

    // copy data to data structures for intersection test
    for (int k=0; k<3; k++){
      orig[k] = ray.orig[k];
      dir[k] = ray.dir[k];
      vert0[k] = vprop->vertex[faceSet->coordIndex[idx+0]][k];
      vert1[k] = vprop->vertex[faceSet->coordIndex[idx+1]][k];
      vert2[k] = vprop->vertex[faceSet->coordIndex[idx+2]][k];
    }

    // call intersection detection routine
    if (intersect_triangle(orig, dir, vert0, vert1, vert2, &t, &u, &v)
	&& t>0.0001 && t<last_t){

      last_t = t;
      hit_idx = idx;
      hit_t = t;
    }

    // end triangle search
  }

  if (visRay)
    lineSet->draw1(ray.orig,
		   ray.orig+ray.dir*((hit_idx>=0)?hit_t:10.0),
		   SbColor(1.0, 1.0, 1.0));
  
  if (hit_idx>=0){
    hit_vtx0 = hit_vtx = faceSet->coordIndex[hit_idx];
    hit_vtx1 = faceSet->coordIndex[hit_idx+1];
    hit_vtx2 = faceSet->coordIndex[hit_idx+2];
    
    // absorb
    if (drand48()>0.5){
      pointSet->add(ray.orig+ray.dir*hit_t);

      // change color
      intens.set1Value(hit_vtx0,
		       intens[hit_vtx0] + SbVec3f(5.0, 5.0, 5.0));
      intens.set1Value(hit_vtx1,
		       intens[hit_vtx1] + SbVec3f(5.0, 5.0, 5.0));
      intens.set1Value(hit_vtx2,
		       intens[hit_vtx2] + SbVec3f(5.0, 5.0, 5.0));
      faceMat->diffuseColor.set1Value(hit_vtx0,
                                      nlscale(intens[hit_vtx0]));
      faceMat->diffuseColor.set1Value(hit_vtx1,
                                      nlscale(intens[hit_vtx1]));
      faceMat->diffuseColor.set1Value(hit_vtx2,
                                      nlscale(intens[hit_vtx2]));
    }

    // reflect
    else {
      rayType ray2;
      ray2.orig = ray.orig+ray.dir*hit_t;
      
      // pick a direction to fire
      sbr.setValue(SbVec3f(0.0, 0.0, 1.0),
		   vprop->normal[hit_vtx]);
      phi = drand48() * 2.0 * M_PI;
      theta = acos(sqrt(drand48()));;
      dp.setValue(sin(theta)*cos(phi),
		  sin(theta)*sin(phi),
		  cos(theta));
      sbr.multVec(dp, dpn);
      dpn.normalize();
      ray2.dir = dpn;
      
      fireray(ray2, visRay);
    }

  }

}






















void
animate(void *, SoSensor *)
{
  
  int vertex = 0;
  int i;
  SbVec3f dp, dpn;
  SbRotation sbr;
  float dot1, dot2;
  SbVec3f edge1, edge2;
  double phi, theta;
  rayType ray;
  static int frame = 0;

  // write scene
  if (! ((frame++)%100)){
    fprintf(stderr, "\rWriting file scene%d.iv, %d photons in scene.",
	    frame/100, pointSet->getNum());
    writeScene();
  }

  sbr.setValue(SbVec3f(0.0, 0.0, 1.0), vprop->normal[firstLightVert]);

  // loop through vertices
  for (vertex=0; vertex<verticesPerPass; vertex++){

    // pick a source
    phi = drand48();
    theta = drand48();
    edge1 =       phi*vprop->vertex[firstLightVert] +
            (1.0-phi)*vprop->vertex[firstLightVert+1];
    edge2 =       theta*vprop->vertex[firstLightVert] +
            (1.0-theta)*vprop->vertex[firstLightVert+3];
    ray.orig = edge1-vprop->vertex[firstLightVert]+edge2;
    
    // shoot a bunch of rays
    for (i=0; i<raysPerVertex; i++){

      // pick a direction
      phi = drand48() * 2.0 * M_PI;
      theta = acos(sqrt(drand48()));;
      dp.setValue(sin(theta)*cos(phi),
		  sin(theta)*sin(phi),
		  cos(theta));
      sbr.multVec(dp, dpn);
      dpn.normalize();
      ray.dir = dpn;

      fireray(ray, (i==raysPerVertex-1)?1:0);
    }
  }

  return;
}

////////////////////////////////////////////////////////////
//
// Much of this code is from Dr. David C. Banks
// http://www.cs.fsu.edu/~banks/
// 
////////////////////////////////////////////////////////////
//
// The main program gets a window from X, 
// puts an interactive 3D examinerViewer in it, 
// attaches the scene graph to the viewer,
// renders the scene graph,
// and hands control over to the Inventor main loop.

void
main(int argc, char* argv[])
{

  // read command line arguments
  int i;
  for (i=1; i<argc; i++){
    if (strcmp(argv[i], "-g") == 0)
      myGamma = atof(argv[i+1]);
    else if (strcmp(argv[i], "-b") == 0){
      fprintf(stderr, "Using brain.\n");
      useBrain = 1;
    }
    else if (strcmp(argv[i], "-nw") == 0){
      fprintf(stderr, "Not creating any windows.\n");
      useWindow = 0;
    }
    else if (strcmp(argv[i], "-h") == 0){
      cerr << "\nUsage: " << argv[0]
	   << " [ -h ] [ -nw ] [ -b ]\n\n"
	   << " -b     Use brain model (requires brain2 dataset)\n"
           << " -h     This help message\n"
	   << " -nw    Run in background (No Windows)\n"
	 ;
      exit(1);
    }
  }

    
  // Initialize new classes
  srand48(665);

  // Create a scene graph
  SoDB::init();
  //SoXt::init(argv[0]);
  root = makeScene();
  root->ref();
   
  if (useWindow){
    Widget myWindow = SoXt::init(argv[0]);
    if (myWindow == NULL) exit(1);
    SoTimerSensor *sensor = new SoTimerSensor(animate, root);
    sensor->setInterval(SbTime(0.015)); // 70 frames per second
    sensor->schedule();
  
    SoXtExaminerViewer *myViewer = new SoXtExaminerViewer(myWindow);
    myViewer->setSceneGraph(root);
    Camera = myViewer->getCamera();
    myViewer->setTitle("Photon Mapping");
    myViewer->setTransparencyType(SoGLRenderAction::SORTED_OBJECT_BLEND);
    myViewer->setBackgroundColor(SbColor(0.15, 0.035, 0.15));
    myViewer->show();

    SoXt::show(myWindow);
    SoXt::mainLoop();

  }
  else {
    fprintf(stderr, "Starting loop...\n");
    while(1){
      animate((void *) NULL, (SoSensor *) NULL);
    }
  }
}
